from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime
from typing import List, Any

from ..models import ValueModel

@dataclass
class ValueResponseModel:
    IsOffline: bool
    OnlineImage: str
    Date: datetime
    FuzzyDate: str
    Values: List[ValueModel]

@dataclass
class HistoryResponseModel:
    label: str
    data: List[List[Any]]
    color: str
    variableid: int
    unit: str
    isfullzoom: bool
    reloadoverview: bool
    yaxis: int
    NumberOfDecimals: int


from __future__ import annotations

import requests
import re

from typing import List
from datetime import datetime

from ..config import MyUpwayConfig
from ..exceptions import LoginErr, NotLoggedIn, ResponseError
from ..models import ValueModel, VariableHistoryValue, VariableValue
from ..enums import Variable
from ..responses import ValueResponseModel, HistoryResponseModel

TIME_REPRESENTATION_FORMATS = [
    "%m/%d/%Y %H:%M:%S",
    "%Y. %m. %d. %H:%M:%S",
    "%Y-%m-%d %H:%M:%S",
    "%d-%m-%Y %H:%M:%S",
    "%d.%m.%Y %H:%M:%S",
    "%d.%m.%Y %H.%M.%S"
]


class MyUpwayService:
    _DOMAIN = 'myupway.com'
    _BASE_URL = f'https://www.{_DOMAIN}'

    _config: MyUpwayConfig
    _session: requests.Session

    deviceId: str
    isOnline: bool

    def __init__(self, config: MyUpwayConfig) -> None:
        self._config = config
        self._session = requests.Session()
        # Set language to en to be able to get boolean values right
        self._session.cookies.set("EmilLanguage", "en-GB", domain=self._DOMAIN)

        self.deviceId = self._config.heatpump_id

        self.login()

    def login(self) -> None:

        url = self._BASE_URL + '/LogIn'
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        data = {
            'returnUrl': '',
            'Email': self._config.username,
            'Password': self._config.password
        }

        response = self._session.post(url, headers=headers, data=data)

        if response.status_code != 200:
            raise requests.exceptions.HTTPError(
                f"Request failed with status code {response.status_code}")

        if not '.ASPXAUTH' in self._session.cookies:
            raise LoginErr("Login failed.")

        self.get_current_values() # perform request to set the isOnline flag

    def get_current_values(self, variables: List[Variable] | None = None, force_login: bool = False) -> List[VariableValue]:
        """
        Returns current values for requested variables provided as list of VariableValue.
        If variables are not specified, function returns all variables.
        """

        url = self._BASE_URL + '/PrivateAPI/Values'
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        data = [
            ('hpid', self._config.heatpump_id)
        ]

        if not variables:
            variables = list(Variable)

        for variable in variables:
            data.append(('variables', variable.value))  # type: ignore

        if not '.ASPXAUTH' in self._session.cookies and force_login:
            self.login()

        if not '.ASPXAUTH' in self._session.cookies:
            raise NotLoggedIn(
                "Session is not currenly logged in. Use force_login = True if you want to force relogin.")

        response = self._session.post(url, headers=headers, data=data)

        if response.status_code != 200:
            raise requests.exceptions.HTTPError(
                f"Request failed with status code {response.status_code}, data: {response}")

        try:
            response_data = response.json()
        except requests.exceptions.JSONDecodeError:
            raise ResponseError(
                f"Cannot parse response data. Data: {response.content}")

        for timeformat in TIME_REPRESENTATION_FORMATS:
            try:
                response_data['Date'] = datetime.strptime(
                    response_data['Date'], timeformat)

            except ValueError:
                continue

            else:
                break

        currentValues = ValueResponseModel(**response_data)

        self.isOnline = not currentValues.IsOffline

        results = []

        for value in currentValues.Values:
            value = ValueModel(**value)  # type: ignore
            id = value.VariableId
            enumerator = Variable(id)
            name = enumerator.name

            # Regular expression pattern to match numeric value and unit
            pattern = r'^(-?[\d]+(?:[.,]\d{1,})?)(\D*)$'

            match = re.match(pattern, value.CurrentValue)
            if match:
                value = match.group(1)  # Extract the numeric value
                unit = match.group(2)  # Extract the unit
                if unit == "":
                    unit = None

            else:
                # If no match, consider the entire value as the value itself
                value = value.CurrentValue
                unit = None

            results.append(VariableValue(id, name, enumerator, value, unit))

        return results

    def get_history_values(self, variable: Variable, startDate: datetime, stopDate: datetime, resolution: int = 1000, force_login: bool = False) -> List[VariableHistoryValue]:
        """
        Returns history values for selected variable from specified timerange.
        """

        url = self._BASE_URL + '/PrivateAPI/History'

        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        data = {
            'hpid': self._config.heatpump_id,
            'variableId': variable.value,
            'resolution': resolution,
            'startDate': int(startDate.timestamp()),
            'stopDate': int(stopDate.timestamp()),
            'isFullZoom': True,
            'reloadOverview': True
        }

        if not '.ASPXAUTH' in self._session.cookies and force_login:
            self.login()

        if not '.ASPXAUTH' in self._session.cookies:
            raise NotLoggedIn(
                "Session is not currenly logged in. Use force_login = True if you want to force relogin.")

        response = self._session.post(url, headers=headers, data=data)

        if response.status_code != 200:
            raise requests.exceptions.HTTPError(
                f"Request failed with status code {response.status_code}, data: {response}")

        try:
            response_data = HistoryResponseModel(**response.json())

        except requests.exceptions.JSONDecodeError:
            raise ResponseError(
                f"Cannot parse response data. Data: {response.content}")

        results = []

        for value in response_data.data:
            history_value = value[1]
            history_date = datetime.fromtimestamp(value[0]/1000)

            if response_data.unit == "":
                history_unit = None
            else:
                history_unit = response_data.unit

            results.append(VariableHistoryValue(
                history_value, history_unit, history_date))

        return results

    def logout(self):
        url = self._BASE_URL + '/LogOut'

        response = self._session.get(url)



from __future__ import annotations

import requests

from typing import List
from datetime import datetime

from ..config import MyUpwayConfig
from ..exceptions import LoginErr
from ..enums import Variable
from ..models import VariableValue, VariableHistoryValue

class MyUplinkService:
    _BASE_URL = 'https://api.myuplink.com'
    
    _config: MyUpwayConfig
    _session: requests.Session
    _token: str

    deviceId: str
    isOnline: bool

    def __init__(self, config: MyUpwayConfig) -> None:
        self._config = config
        self._session = requests.Session()

        self.login()
    
    def login(self):
        self._get_token()
    
        response = self._session.get(self._BASE_URL + '/v2/systems/me?page=1&itemsPerPage=10')
        response_data = response.json()

        device_data = response_data['systems'][0]['devices'][0]

        self.deviceId = device_data['id']
        self.isOnline = device_data['connectionState'] == "Connected"

    def get_current_values(self, variables: List[Variable] | None = None, force_login: bool = False) -> List[VariableValue]:
        """
        Returns current values for requested variables provided as list of VariableValue.
        If variables are not specified, function returns all variables.
        """

        params = {}

        if variables:
            params = {
                "parameters": ','.join(str(variable.value) for variable in variables)
            }
        
        response = self._session.get(self._BASE_URL + '/v2/devices/' + self.deviceId + '/points', params=params)

        if response.status_code == 401:
            self._get_token()

            response = self._session.get(self._BASE_URL + '/v2/devices/' + self.deviceId + '/points', params=params)

        elif response.status_code != 200:
            raise Exception(f"Failed to get values. API responded with status code {response.status_code}")

        response_data = response.json()

        results: List[VariableValue] = []

        for result in response_data:
            try:
                variableEnum = Variable(int(result["parameterId"]))
            except ValueError:
                print(f"ERROR: Missing Variable enum. Contact the maintainer (https://github.com/lemanjo/pyupway) to add it to the code with these details: {result}")
                continue

            raw_value = result["value"]
            # JSON does not distinguish between int and float, so python always outputs float. But enum
            # interpretation needs to be able to access the value as an int if it is one.
            value = int(raw_value) if raw_value.is_integer() else raw_value
            str_value = str(value)

            found_enum = [el for el in result["enumValues"] if str(el["value"]) == str_value]
            enum_value = found_enum[0]["text"] if len(found_enum) > 0 and "text" in found_enum[0] else str_value

            variable_value = VariableValue(
                Id=variableEnum.value,
                Name=variableEnum.name,
                Value=raw_value,
                Unit=result["parameterUnit"],
                EnumValue=enum_value,
                Enumerator=variableEnum
            )

            if not self._contains_variable_value_with_id(results, variable_value.Id):
                results.append(variable_value)
        
        return results

    def get_history_values(self, variable: Variable, startDate: datetime, stopDate: datetime, resolution: int = 1000, force_login: bool = False) -> List[VariableHistoryValue]:
        """
        No way to implement in new API.
        """

        results: List[VariableHistoryValue] = []

        return results
    
    def logout(self):
        """ No implementation """
        pass

    
    def _get_token(self):
        """
        Fetch new token from OAUTH
        """
        
        # Define token request parameters
        token_params = {
            'grant_type': 'client_credentials',  # or other grant type depending on your OAuth flow
            'client_id': self._config.client_id,
            'client_secret': self._config.client_secret,
            # additional parameters if needed
        }

        # Make a POST request to the token endpoint
        response = requests.post(self._BASE_URL+'/oauth/token', data=token_params)

        # Check if the request was successful (status code 200)
        if response.status_code == 200:
            # Extract the token from the response
            token_data = response.json()
            access_token = "Bearer "+token_data['access_token']
            self._token = access_token
            self._session.headers.update({'Authorization': access_token})
        else:
            raise LoginErr("Cannot fetch MyUplink token")
    
    def _contains_variable_value_with_id(self, variable_values, Id):
        for var_value in variable_values:
            if var_value.Id == Id:
                return True
        return False



from __future__ import annotations
from dataclasses import dataclass
from ..enums import Variable

@dataclass
class VariableValue:
    Id: int
    Name: str
    Enumerator: Variable
    Value: int | float | str | bool | None
    Unit: str | None
    EnumValue: str | None



from __future__ import annotations
from typing import Optional
from ..enums import DataService
from ..exceptions import ConfigurationError


class MyUpwayConfig:
    username: Optional[str]
    password: Optional[str]
    heatpump_id: Optional[int]
    dataservice: DataService
    client_id: Optional[str]
    client_secret: Optional[str]

    def __init__(self, dataservice: DataService, username: str = None, password: str = None, heatpump_id: int = None, client_id: str = None, client_secret: str = None) -> None:
        self.dataservice = dataservice

        if self.dataservice == DataService.MYUPWAY:
            if not username:
                raise ConfigurationError("Username needs to be defined when using MyUpway")
            self.username = username

            if not password:
                raise ConfigurationError("Password needs to be defined when using MyUpway")
            self.password = password

            if not heatpump_id:
                raise ConfigurationError("Heatpump ID needs to be defined when using MyUpway")
            self.heatpump_id = heatpump_id
        
        if self.dataservice == DataService.MYUPLINK:
            if not client_id:
                raise ConfigurationError("Client id configuration is required for MyUplink")
            self.client_id = client_id

            if not client_secret:
                raise ConfigurationError("Client secret configuration is required for MyUplink")
            self.client_secret = client_secret


from __future__ import annotations
from dataclasses import dataclass
from datetime import datetime

@dataclass
class VariableHistoryValue:
    Value: int | float | str | bool | None
    Unit: str | None
    Date: datetime



from __future__ import annotations
from typing import List
from datetime import datetime

from .enums import DataService, Variable
from .config import MyUpwayConfig
from .models import VariableValue, VariableHistoryValue
from .services import MyUpwayService, MyUplinkService


class MyUpway:
    _config: MyUpwayConfig

    deviceId: str
    isOnline: bool

    def __init__(self, config: MyUpwayConfig) -> None:
        self._config = config

        if self._config.dataservice == DataService.MYUPWAY:
            self._service = MyUpwayService(self._config)
        else:
            self._service = MyUplinkService(self._config)

        self._service.login()

        self.deviceId = self._service.deviceId
        self.isOnline = self._service.isOnline

    def login(self):
        self._service.login()

    def get_current_values(self, variables: List[Variable] | None = None, force_login: bool = False) -> List[VariableValue]:
        """
        Returns current values for requested variables provided as list of VariableValue.
        If variables are not specified, function returns all variables.
        """

        return self._service.get_current_values(variables, force_login)

    def get_history_values(self, variable: Variable, startDate: datetime, stopDate: datetime, resolution: int = 1000, force_login: bool = False) -> List[VariableHistoryValue]:
        """
        MyUpway: Returns history values for selected variable from specified timerange.
        MyUplink: Not implemented. History data not available through new API.
        """

        return self._service.get_history_values(variable, startDate, stopDate, resolution, force_login)

    def logout(self):
        """
        MyUpway: Logs you out from the system.
        MyUplink: Not implemented as integration is token based API.
        """

        self._service.logout()



class LoginErr(Exception):
    def __init__(self, message="Cannot login"):
        self.message = message
        super().__init__(self.message)

class NotLoggedIn(Exception):
    def __init__(self, message="Not logged in"):
        self.message = message
        super().__init__(self.message)


from enum import Enum


class Variable(Enum):
    """
    Provides Variable name - Variable ID mapping as enum
    """

    AVG_OUTDOOR_TEMP = 40067
    HOT_WATER_CHARGING = 40014                  # TehoWatti AIR
    HOT_WATER_TOP = 40013                       # TehoWatti AIR
    INDOOR_UNIT_OUTDOOR_TEMP = 40004
    CURRENT_BE1 = 40083
    CURRENT_BE2 = 40081
    CURRENT_BE3 = 40079
    MYUPWAY_DEGREE_MINUTES = 43005
    EXTERNAL_ADJUSTMENT = 43161
    FLOOR_DRYING_FUNCTION = 47276
    CALCULATED_FLOW_TEMP = 43009
    EXTERNAL_FLOW_TEMP = 40071
    EXTERNAL_RETURN_TEMP = 40152
    HEAT_MEDIUM_FLOW = 40008                    # TehoWatti AIR
    HEAT_RETURN_TEMP = 40012                    # TehoWatti AIR
    ROOM_TEMPERATURE = 40033
    ADDITION_BLOCKED = 10033
    ADDITION_MAX_STEP = 47613
    ADDITION_STATUS = 43091
    ADDITION_FUSE_SIZE = 47214
    ADDITION_TIME_FACTOR = 43081                # TehoWatti AIR
    ADDITION_ELECTRICAL_ADDITION_POWER = 43084  # TehoWatti AIR
    ADDITION_SET_MAX_ELECTRICAL_ADD = 47212     # TehoWatti AIR
    ADDITION_TEMPERATURE = 40121                # TehoWatti AIR
    ENERGY_COOLING_COMPRESSOR_ONLY = 44302      # TehoWatti AIR
    ENERGY_HEATING_COMPRESSOR_ONLY = 44308      # TehoWatti AIR
    ENERGY_HEATING_INT_ADD_INCL = 44300         # TehoWatti AIR
    ENERGY_HOTWATER_COMPRESSOR_ONLY = 44306     # TehoWatti AIR
    ENERGY_HW_INCL_INT_ADD = 44298              # TehoWatti AIR
    ENERGY_POOL_COMPRESSOR_ONLY = 44304         # TehoWatti AIR
    ENERGY_FLOW = 40072                         # TehoWatti AIR
    INVERTER_STATUS = 50096
    AUX1 = 47411
    AUX2 = 47410
    AUX3 = 47409
    AUX4 = 47408
    AUX5 = 47407
    AUX6 = 48366
    X7 = 47412
    COUNTRY = 48745
    DEFROSTING = 44703
    CHARGE_PUMP_SPEED = 44396
    OUTDOOR_UNIT_OUTDOOR_TEMP = 44362
    COMPRESSOR_BLOCKED = 10014
    COMPRESSOR_STARTS = 44069
    COMPRESSOR_PROTECTION_MODE = 44702
    CONDENSER_OUT = 44058
    EVAPORATOR = 44363
    HOT_GAS = 44059
    LIQUID_LINE = 44060
    RETURN_TEMP = 44055
    SUCTION_GAS = 44061
    HIGH_PRESSURE_SENSOR = 44699
    HIGH_PRESSURE_SENSOR_EB101 = 41163
    LOW_PRESSURE_SENSOR = 44700
    COMPRESSOR_OPERATING_TIME = 44071
    COMPRESSOR_OPERATING_TIME_HOT_WATER = 44073
    COMPRESSOR_RUN_TIME_COOLING = 40737
    CURRENT_COMPRESSOR_FREQUENCY = 44701
    REQUESTED_COMPRESSOR_FREQUENCY = 40782
    VERSION = 44014
    SMART_PRICE_STATUS = 44908          # Metro-air 330
    SMART_PRICE_VALUE = 10069             # Metro-air 330
    SMART_PRICE_FACTOR = 44896          # Metro-air 330
    SUPPLY_LINE = 40047                 # MYUPLINK
    RETURN_LINE = 40048                 # MYUPLINK
    OIL_TEMPERATURE_EP15_BT29 = 40145   # MYUPLINK
    OIL_TEMPERATURE_BT29 = 40146        # MYUPLINK
    MYUPLINK_DEGREE_MINUTES = 40940     # MYUPLINK
    SLAVE_EB101 = 44032                 # MYUPLINK
    STATUS_COMPRESSOR_EB101 = 44064     # MYUPLINK
    CHARGE_PUMP_EB101_GP12 = 49996      # MYUPLINK
    CURRENT_HOT_WATER_MODE = 43109
    HOT_WATER_CHARGE_SET_POINT_VALUE = 43115
    PUMP_HEATING_MEDIUM = 49992
    INT_ELEC_ADD_HEAT = 49993
    PRIORITY = 49994
    PUMP_HEATING_MEDIUM_2 = 49995
    TEMPORARY_LUX = 50004
    COMFORT_MODE = 47041
    COMPRESSOR_EVAPORATOR_TEMP = 40020          # NIBE F470
    COMPRESSOR_TEMP = 40023                     # NIBE F470
    ADDITION_IMMERSION_HEATER_TEMP = 40024      # NIBE F470
    AIR_EXHAUST_TEMP = 40025                    # NIBE F470
    AIR_EXTRACT_TEMP = 40026                    # NIBE F470
    # NIBE F470  # ENUM: 0=unknown, 1=low, 2=normal, 3=high
    SMART_PRICE_MODE = 41929
    DESIRED_HUMIDITY_1 = 42770                  # NIBE F470
    DEFROSTING_TIME = 43066                     # NIBE F470
    FAN_MODE = 43108                            # NIBE F470
    # NIBE F470  # ENUM: 20=off, 40=starts, 60=runs, 80=stops
    COMPRESSOR_STATUS = 43427
    SMART_PRICE_POOL_OFFSET = 44898             # NIBE F470
    SMART_PRICE_COOLING_OFFSET = 44899          # NIBE F470
    # NIBE F470  # ENUM: 0=auto, 1=manual, 2=addition_only
    OPERATING_MODE = 47137
    DESIRED_HUMIDITY_2 = 49633                  # NIBE F470
    AIR_INCREASED_VENTILATION = 50005           # NIBE F470  # ENUM: 0=off, 1=on
